# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Virtual gallery constants.
"""

import kapture

virtual_gallery_camera_model = kapture.CameraType.PINHOLE
virtual_gallery_width = 1920
virtual_gallery_height = 1080
