# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Additional utility functions and classes used by kapture but not specific to it.
"""

from .Collections import AutoEnum  # noqa: F401
from .paths import path_secure, populate_files_in_dirpath  # noqa: F401
